#!/usr/local/bin/perl -i~

#THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT
#LAWS OF THE UNITED STATES.  FOR LICENSING INFORMATION CONTACT:
#
#Christian Bischof or Lucas Roh, Mathematics and Computer Science Division,    
#Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL 60439, 
#{bischof,roh}@mcs.anl.gov.

require 5;

die "Need to set ADIC environment variable.\n" if ( ! defined ($ENV{'ADIC'}) );
unshift(@INC,"$ENV{'ADIC'}/bin");
require 'commabreak.pl';
require 'newgetopt.pl';


chop ($progname = `basename $0`);

$usage = "usage: $progname GRADIENT-REPRESENTATION [MODE]
       GRADIENT-REPRESENTATION: One of -array, -pointer, or -nograd.
       MODE: For pointer, one may choose -direct, -inline, or -indirect.
";

$" = ",";
$_ = " " x 2500;

# Used only in Pointer-Inline mode.
$validation = " " x 2000;
$accumulation = " " x 2000;



$error = ! &NGetOpt("array",
                    "macro",
                    "pointer",
                          "direct", "indirect", "inline",
                    "nograd"
                    );
    
die "$usage" if ( $error );

# Forgot to tell us what to do.

die "$usage" if ! ( $opt_array || $opt_macro );
    
# Selected pointer mode but did not specify accumulation style.

if ( $opt_pointer 
    && ! ( $opt_direct || $opt_indirect || $opt_inline ) )
{
    die "$progname: Must specify -direct, -indirect, or -inline with -pointer.\n";
}

# Wanted -array or -nograd and also specified some accumulation style.

if ( ( $opt_array || $opt_nograd )
    && ( $opt_direct || $opt_indirect || $opt_inline ) )
{
    die "$progname: Cannot specify accumulation style (-direct, -indirect, or -inline) with -array or -nograd\n";
}


if ( $opt_pointer )
{
    $ad_do_localize_gradients = 1;
    $ad_do_authorize_arrays = 1;
}
else
{
    $ad_do_localize_gradients = 0;
    $ad_do_authorize_arrays = 0;
}



    open(INPUT_FILE, "$ARGV[$#ARGV - 1]") || die "Warning: cannot open $file";
    open(OUT_FILE, ">$ARGV[$#ARGV]") || die "Error: cannot open $file for writing.";

while (<INPUT_FILE>)
{
    if ( /DEAUTHORIZE/ )
    {
        @args = &commabreak($_);
        pop(@args);  # Get rid of the \000



        next if ( m/DEAUTHORIZE\(a_/ );

        if ( $ad_do_authorize_arrays )
        {
            if ( $#args == 0 )
            {
                # Only one element = just name; do scalar deauthorization
                $_ = "AD_AUTHORIZE_ONE($args[0]);\n";
            }
            else 
            {
                # More elts = array; do array deauthorization
                # (Array separator interpolating as comma in strings.)
                $_ = 
                    "/* DEAUTHORIZE (@args) */\n" . 
                        "#ifdef AD_DO_AUTHORIZE_ARRAYS\n" .
                            &array_work("AD_DEAUTHORIZE_ONE",@args) .
                                "#endif /* AD_DO_AUTHORIZE_ARRAYS */\n";
            }
        }
        else
        {
            $_ = "";
        }


    }
    elsif ( /AUTHORIZE/ )
    {
        @args = &commabreak($_);
        pop(@args);  # Get rid of the \000



        
        if ( $ad_do_authorize_arrays )
        {
            if ( $#args == 0 )
            {
                # Only one element = just name; do scalar authorization
                $_ = "AD_AUTHORIZE_ONE($args[0]);\n";
            }
            else 
            {
                # More elts = array; do array authorization
                # (Array separator interpolating as comma in strings.)
                $_ = 
                    "/* AUTHORIZE (@args) */\n" . 
                        "#ifdef AD_DO_AUTHORIZE_ARRAYS\n" .
                            &array_work("AD_AUTHORIZE_ONE",@args) .
                                "#endif /* AD_DO_AUTHORIZE_ARRAYS */\n";
            }
        }
        else
        {
            $_ = "";
        }


    } 
    elsif ( /LOCALIZE_GRADIENT/ )
    { 
        @args = &commabreak($_);
        pop(@args);  # Get rid of the \000



        
        if ( $ad_do_localize_gradients )
        {
            if ( $#args == 0 )
            {
                # Only one element = just name; do scalar localization
                $_ = "AD_LOCALIZE_ONE($args[0]);\n";
            }
            else 
            {
                # More elts = array; do array deauthorization
                # (Array separator interpolating as comma in strings.)
                $_ = 
                    "/* LOCALIZE_GRADIENT (@args) */\n" . 
                        "#ifdef AD_DO_LOCALIZE_GRADIENTS\n" .
                            &array_work("AD_LOCALIZE_ONE",@args) .
                                "#endif /* AD_DO_LOCALIZE_GRADIENTS */\n";
            }
        }
        else
        {
            $_ = "";
        }


    }
    elsif ( /AD_INTRINSIC_SETUP_CODE/ )
    {
        @args = &commabreak($_);
        pop(@args);  # Get rid of the \000



        {
            $_ = "
        static int AD_INTRINSIC_FUNCTION_ID = 0;
        if (AD_INTRINSIC_FUNCTION_ID == 0) {
            AD_INTRINSIC_SETUP_FUNCTION_ID(&AD_INTRINSIC_FUNCTION_ID, __FILE__, $args[0]);
        }
        ";

        }
            

    }
    elsif ( /ad_grad_axpy_n/ )
    {
        @args = &commabreak($_);
        pop(@args);  # Get rid of the \000



        
        next if $opt_nograd;

        $arity = $args[0];
        $dest_vector = $args[1];

        @alphas = ();
        @vectors = ();

        for ( $i = 2; $i < 2*$arity + 2 ; $i = $i + 2 )
        {
            push(@alphas, $args[$i]);
            push(@vectors, $args[$i+1]);
        }




        if ( $arity == 0 )
        {
            $_ = "ad_grad_axpy_zero($dest_vector);\n";
        }
        elsif ( $opt_macro )
        {
            



        }
        else
        {
            {
                $_ = "ad_grad_axpy_n($arity,$dest_vector"; 
                for $i ( 0 .. $arity - 1 )
                {
                    $_ .= ",$alphas[$i],$vectors[$i]";
                }
                
                # Append closing semicolon and braces 
                $_ .= ");\n";

            }

        }


    }

    print OUT_FILE;
}


sub array_work {
    local ($i);
    local ($dims);
    local ($DESIRED_MACRO, @args) = @_;
    
    $array_name = $args[0];
    $dims = $#args;  # array dims = 1 .. length of @args 
            
    # Pre-extend $out.
    if ( ! $out_extended )
    {
        $out = " "x2000;
        $out_extended = 1;
    }

    # Begin with open block to make new declarations OK.
    $out = "{"; # }

    # Declare tmp vars
    foreach $i ( 1 .. $dims)
    {
        $out .= "int g_i_$i;\n";
        $out .= "for (g_i_$i = 0; g_i_$i < $args[$i]; g_i_$i++) {\n"; #}
    }

    # Insert macro depending on what the current task is.
    # (e.g., AD_AUTHORIZE_ONE, AD_DEAUTHORIZE_ONE, AD_LOCALIZE_ONE)
    $out .= "$DESIRED_MACRO(${array_name}"; #)
    
    # Add array indices 
    foreach $i ( 1 .. $dims) {
        $out .= "[g_i_$i]";
    }

    # Close DESIRED_MACRO(
    $out .= ");\n";
            
    # Close for loops
    foreach $i ( 1 .. $dims)
    { #{
        $out .= "} ";
    }

    # End with a newline.    {
    $out .= "}\n";

    return ($out);
}



