
#ifndef AD_HESSIAN_H
#define AD_HESSIAN_H

extern int hess_size_1, hess_size_2;
extern int hess_maxsize_1, hess_maxsize_2;
extern int iWiLlNeVeRCoNfLiCt;

#include "hessian.h"

#define ad_AD_HessInit(n)

#define ad_AD_HessFinal() \
{ \
    hess_size_1 = 0; \
    hess_size_2 = 0; \
}

#define ad_AD_ClearGrad(gz)\
{\
   int iWiLlNeVeRCoNfLiCt0;\
   for (iWiLlNeVeRCoNfLiCt0 = 0 ; iWiLlNeVeRCoNfLiCt0 < hess_maxsize_1; \
     iWiLlNeVeRCoNfLiCt0++) {\
                               DERIV_grad(gz)[iWiLlNeVeRCoNfLiCt0] = 0.0;\
   }\
   for (iWiLlNeVeRCoNfLiCt0 = 0 ; iWiLlNeVeRCoNfLiCt0 < hess_maxsize_2; \
     iWiLlNeVeRCoNfLiCt0++) {\
                               DERIV_hess(gz)[iWiLlNeVeRCoNfLiCt0] = 0.0;\
   }\
}

#define ad_AD_GetNumIndependents() hess_size_1
#define ad_AD_SetNumIndependents(x) { hess_size_1 = x}
#define ad_AD_ExtractGrad(a, var) \
{ \
    int pOsItIoN; \
    for (pOsItIoN = 0; pOsItIoN < hess_size_1; pOsItIoN++) {\
            (a)[pOsItIoN] = DERIV_grad(var)[pOsItIoN];  \
    }\
}

#define ad_AD_ExtractHessEntry(a, val, i, j)\
{\
  int pOsItIoN1, pOsItIoN2;\
  pOsItIoN1 = (i) > (j) ? (i) : (j); \
  pOsItIoN2 = (i) < (j) ? (i) : (j); \
  val = (a).hess[pOsItIoN2+pOsItIoN1*(pOsItIoN1+1)/2]; \
}

#define ad_AD_ExtractHess(a,var) \
{ \
  int pOsItIoN1, pOsItIoN2; \
  for (pOsItIoN1 = 0; pOsItIoN1 < hess_size_1; pOsItIoN1++) { \
    for (pOsItIoN2 = 0; pOsItIoN2 < pOsItIoN1+1; pOsItIoN2++) { \
      var[pOsItIoN1][pOsItIoN2] = (a).hess[pOsItIoN2+pOsItIoN1*(pOsItIoN1+1)/2]; \
      var[pOsItIoN2][pOsItIoN1] = var[pOsItIoN1][pOsItIoN2]; \
    } \
  } \
}
  
#define ad_AD_SetGrad(a, var) \
    { \
        int pOsItIoN; \
        for (pOsItIoN = 0; pOsItIoN < hess_size_1; pOsItIoN++) {\
            DERIV_grad(var)[pOsItIoN] = (a)[pOsItIoN];  \
        }\
    }

#define ad_AD_SetIndepDone() \
{ \
  hess_size_2 = hess_size_1*(hess_size_1+1)/2; \
}

#define ad_AD_ResetIndep() \
{ \
  hess_size_1 = 0; \
}

#define ad_AD_SetIndep(var) \
{ \
  int pOsItIoN = hess_size_1++; \
  if (pOsItIoN > HESS_MAXVARS) { \
    fprintf (stderr, "ad_AD_SetIndep() Error: the number of " \
	     "independent variables %d exceeds the max (HESS_MAXVARS)!\n",\
	     pOsItIoN); \
    fflush(stderr);  \
    abort();  \
  }\
  ad_AD_ClearGrad(var); \
  DERIV_grad(var)[pOsItIoN] = 1; \
}

#define ad_AD_SetIndepArray(vars, size) \
{ \
  int iWiLlNeVeRCoNfLiCt; \
  for (iWiLlNeVeRCoNfLiCt = 0; iWiLlNeVeRCoNfLiCt < size; iWiLlNeVeRCoNfLiCt++) { \
    ad_AD_ClearGrad((vars)[iWiLlNeVeRCoNfLiCt]); \
    DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])[hess_size_1++] = 1; \
  } \
}

#define ad_AD_CompleteTaylorMap(map,numvars) \
{ \
  int iWiLlNeVeRCoNfLiCt, iWiLlNeVeRCoNfLiCt2; \
  hess_size_2 = numvars; \
  for (iWiLlNeVeRCoNfLiCt=0; iWiLlNeVeRCoNfLiCt<(numvars); iWiLlNeVeRCoNfLiCt++) { \
    for (iWiLlNeVeRCoNfLiCt2=0; iWiLlNeVeRCoNfLiCt2<(numvars); iWiLlNeVeRCoNfLiCt2++) { \
      if ((map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] == 1) \
	(map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt] = 1; \
      if ((map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] == 1 &&  \
	  (map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt] != -1) \
	(map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt] = 1; \
      if ((map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] == 1 && \
	  (map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt2] != -1) \
	(map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt2] = 1; \
    } \
  } \
  for (iWiLlNeVeRCoNfLiCt=0; iWiLlNeVeRCoNfLiCt<(numvars); iWiLlNeVeRCoNfLiCt++)  \
    for (iWiLlNeVeRCoNfLiCt2=0; iWiLlNeVeRCoNfLiCt2<(numvars); iWiLlNeVeRCoNfLiCt2++)  \
      if ((map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] == -1) \
	(map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] = 0; \
  hess_size_1 = 1; \
  for (iWiLlNeVeRCoNfLiCt=0; iWiLlNeVeRCoNfLiCt<(numvars); iWiLlNeVeRCoNfLiCt++) { \
    for (iWiLlNeVeRCoNfLiCt2=0; iWiLlNeVeRCoNfLiCt2<(numvars); iWiLlNeVeRCoNfLiCt2++){ \
      if ((map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] == 1) { \
	(map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] = hess_size_1; \
	(map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt] = hess_size_1; \
	hess_size_1 = hess_size_1 + 1; \
      } \
    } \
  } \
  hess_size_1 = hess_size_1 - 1; \
}

#define ad_AD_InitTaylor(n, map, vec) \
{ \
  int iWiLlNeVeRCoNfLiCt; \
  for (iWiLlNeVeRCoNfLiCt=0; iWiLlNeVeRCoNfLiCt<hess_size_1; iWiLlNeVeRCoNfLiCt++) \
    (vec).taylor[0][iWiLlNeVeRCoNfLiCt] = 0.0; \
  for (iWiLlNeVeRCoNfLiCt=0; iWiLlNeVeRCoNfLiCt<hess_size_1; iWiLlNeVeRCoNfLiCt++) \
    (vec).taylor[1][iWiLlNeVeRCoNfLiCt] = 0.0; \
  for (iWiLlNeVeRCoNfLiCt=0; iWiLlNeVeRCoNfLiCt<hess_size_2; iWiLlNeVeRCoNfLiCt++) { \
    if((map)[iWiLlNeVeRCoNfLiCt][(n)] != 0) \
      (vec).taylor[0][(map)[iWiLlNeVeRCoNfLiCt][(n)]-1] = 1.0; \
  } \
}

#define ad_AD_InterpTaylor(map, vec, grad, hess) \
{ \
  int iWiLlNeVeRCoNfLiCt, iWiLlNeVeRCoNfLiCt2; \
  for (iWiLlNeVeRCoNfLiCt=0; iWiLlNeVeRCoNfLiCt<hess_size_2; iWiLlNeVeRCoNfLiCt++) \
    (grad)[iWiLlNeVeRCoNfLiCt] = 0.0; \
  for (iWiLlNeVeRCoNfLiCt=0; iWiLlNeVeRCoNfLiCt<hess_size_2; iWiLlNeVeRCoNfLiCt++) \
    for (iWiLlNeVeRCoNfLiCt2=0; iWiLlNeVeRCoNfLiCt2<hess_size_2; iWiLlNeVeRCoNfLiCt2++) \
      (hess)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] = 0.0; \
  for (iWiLlNeVeRCoNfLiCt=0; iWiLlNeVeRCoNfLiCt<hess_size_2; iWiLlNeVeRCoNfLiCt++) { \
    if ((map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt] != 0) \
      (grad)[iWiLlNeVeRCoNfLiCt] = \
	(vec).taylor[0][(map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt]-1];   \
    for (iWiLlNeVeRCoNfLiCt2=0; iWiLlNeVeRCoNfLiCt2<iWiLlNeVeRCoNfLiCt; \
	 iWiLlNeVeRCoNfLiCt2++) { \
      if ((map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt] != 0) { \
	if (iWiLlNeVeRCoNfLiCt == iWiLlNeVeRCoNfLiCt2) \
	  hess[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt] = \
	    2.0*(vec).taylor[1][(map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt]-1];  \
	else { \
	  if ((map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt] != 0 && \
	      (map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt2] != 0)  \
	    hess[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] = \
	      (vec).taylor[1][(map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt]-1] -  \
	      (vec).taylor[1][(map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt]-1] -  \
	      (vec).taylor[1][(map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt2]-1]; \
	  else if ((map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt] != 0)  \
	    hess[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt2] = \
	      (vec).taylor[1][(map)[iWiLlNeVeRCoNfLiCt2][iWiLlNeVeRCoNfLiCt]-1] -  \
	      (vec).taylor[1][(map)[iWiLlNeVeRCoNfLiCt][iWiLlNeVeRCoNfLiCt]-1]; 














#endif
