#include "ad_mpi.h"

typedef struct {
        double val;
        int loc;
} MPI_Double_int;

int ad_MPI_Allreduce(void *in, void *out, int count, MPI_Datatype datatype,
		     MPI_Op op, MPI_Comm comm) {
int ierr;

if (datatype != MPI_DERIV_TYPE) { 
  return MPI_Allreduce(in,out,count,datatype,op,comm);
} else {
  /* Need actual code eventually, but this will work for now */
  ierr = ad_MPI_Reduce(in,out,count,datatype,op,0,comm);
  ierr += MPI_Bcast(out,count,datatype,0,comm);
  return ierr;
}
}

int ad_MPI_Reduce(void *in, void *out, int count, MPI_Datatype datatype,
		  MPI_Op op, int root, MPI_Comm comm) {

int i,g_i,pos,where;
int ierr,np,myid,partner;
double bigtemp[4096],bigtemp2[4096];
MPI_Double_int *mtemp1,*mtemp2;
double *temp1,*temp2,*temp3,*temp4;
double prtl,value;
DERIV_TYPE *ad_in,*ad_out;
MPI_Status status;

/* bigtemps should probably be dynamically allocated */
  
  if (datatype != MPI_DERIV_TYPE) { 
    return MPI_Reduce(in,out,count,datatype,op,root,comm);
  } else {
    ad_in = (DERIV_TYPE *) in;
    ad_out = (DERIV_TYPE *) out;
    if (op == MPI_SUM) {
   for(i=0;i<count;i++){
     bigtemp[i] = DERIV_val(ad_in[i]);
   }
   pos = count;

   for(i=0;i<count;i++)
     for(g_i=0;g_i<ad_GRAD_MAX;g_i++)
       bigtemp[pos++] = DERIV_grad(ad_in[i])[g_i];

   ierr = MPI_Reduce(bigtemp,bigtemp2,pos,MPI_DOUBLE,MPI_SUM,root,comm);

   ierr = MPI_Comm_rank(MPI_COMM_WORLD,&myid);

   if (myid==root) {
     for(i=0;i<count;i++)
       DERIV_val(ad_out[i]) = bigtemp2[i];

     pos = count;

     for(i=0;i<count;i++)
       for(g_i=0;g_i<ad_GRAD_MAX;g_i++)
	 DERIV_grad(ad_out[i])[g_i] = bigtemp2[pos++];
   }
    } else if (op == MPI_MAX) {
   ierr = MPI_Comm_rank(MPI_COMM_WORLD,&myid);
   ierr = MPI_Comm_size(MPI_COMM_WORLD,&np);

   mtemp1 = (MPI_Double_int *) calloc(count,sizeof(MPI_Double_int));
   mtemp2 = (MPI_Double_int *) calloc(count,sizeof(MPI_Double_int));

   for(i=0;i<count;i++){
     (mtemp1[i]).val = DERIV_val(ad_in[i]);
     (mtemp1[i]).loc = myid;
   }

   ierr = MPI_Allreduce(mtemp1,mtemp2,count,MPI_DOUBLE_INT,MPI_MAXLOC,comm);
   for(i=0;i<count;i++){
     where = (mtemp2[i]).loc;
     value = (mtemp2[i]).val;
     if (where == myid) {
       /* At some point, it becomes more economical for every process to send
          all of ad_in and let the root sort through it all.  Perhaps some
          rule for when count >> np? */
       ierr = MPI_Send(&ad_in[i],1,MPI_DERIV_TYPE,root,1970,comm);
     } else {
       if (DERIV_val(ad_in[i]) == value)
	 printf("MAX exception!\n");
     }
     if (myid == root)
       ierr = MPI_Recv(&ad_out[i],1,MPI_DERIV_TYPE,where,1970,comm,&status);
   }
    } else if (op == MPI_MIN) {
   ierr = MPI_Comm_rank(MPI_COMM_WORLD,&myid);
   ierr = MPI_Comm_size(MPI_COMM_WORLD,&np);

   mtemp1 = (MPI_Double_int *) calloc(count,sizeof(MPI_Double_int));
   mtemp2 = (MPI_Double_int *) calloc(count,sizeof(MPI_Double_int));

   for(i=0;i<count;i++){
     (mtemp1[i]).val = DERIV_val(ad_in[i]);
     (mtemp1[i]).loc = myid;
   }

   ierr = MPI_Allreduce(mtemp1,mtemp2,count,MPI_DOUBLE_INT,MPI_MINLOC,comm);
   for(i=0;i<count;i++){
     where = (mtemp2[i]).loc;
     value = (mtemp2[i]).val;
     if (where == myid) {
       /* At some point, it becomes more economical for every process to send
          all of ad_in and let the root sort through it all.  Perhaps some
          rule for when count >> np? */
       ierr = MPI_Send(&ad_in[i],1,MPI_DERIV_TYPE,root,1970,comm);
     } else {
       if (DERIV_val(ad_in[i]) == value)
	 printf("MIN exception!\n");
     }
     if (myid == root)
       ierr = MPI_Recv(&ad_out[i],1,MPI_DERIV_TYPE,where,1970,comm,&status);
   }
    } else if (op == MPI_PROD) {
   ierr = MPI_Comm_rank(MPI_COMM_WORLD,&myid);
   ierr = MPI_Comm_size(MPI_COMM_WORLD,&np);

   partner = np - myid - 1;

   pos = 0;

   temp1 = (double *) calloc(count,sizeof(double));
   temp2 = (double *) calloc(count,sizeof(double));
   temp3 = (double *) calloc(count,sizeof(double));
   temp4 = (double *) calloc(count,sizeof(double));

   for(i=0;i<count;i++)
     temp1[i] = DERIV_val(ad_in[i]);

/* Might be able to improve performance using nonblocking communications; as
   implemented, might even be a danger of deadlock on some MPI systems */

   ierr = MPI_Send(temp1,count,MPI_DOUBLE,partner,1988,comm);
   ierr = MPI_Recv(temp2,count,MPI_DOUBLE,partner,1988,comm,&status);
   ierr = MPI_Scan(temp1,temp3,count,MPI_DOUBLE,MPI_PROD,comm);
   ierr = MPI_Scan(temp2,temp4,count,MPI_DOUBLE,MPI_PROD,comm);

   if (myid == (np-1)) {
     ierr = MPI_Send(temp3,count,MPI_DOUBLE,root,1970,comm);
     ierr = MPI_Recv(temp3,count,MPI_DOUBLE,myid-1,1991,comm,&status);
     for(i=0;i<count;i++) {
       prtl = temp3[i];
       /* Really shouldn't use ad_GRAD_MAX; should use adaxpy, if possible */
       for(g_i=0;g_i<ad_GRAD_MAX;g_i++)
	 bigtemp[pos++] = prtl*DERIV_grad(ad_in[i])[g_i];
     }
   } else {
     ierr = MPI_Send(temp3,count,MPI_DOUBLE,myid+1,1991,comm);
     ierr = MPI_Send(temp4,count,MPI_DOUBLE,partner-1,1993,comm);
     ierr = MPI_Recv(temp4,count,MPI_DOUBLE,partner-1,1993,comm,&status);
     if (myid) {
       ierr = MPI_Recv(temp3,count,MPI_DOUBLE,myid-1,1991,comm,&status);
       for(i=0;i<count;i++) {
	 prtl = temp3[i]*temp4[i];
	 /* Really shouldn't use ad_GRAD_MAX; should use adaxpy, if possible */
	 for(g_i=0;g_i<ad_GRAD_MAX;g_i++)
	   bigtemp[pos++] = prtl*DERIV_grad(ad_in[i])[g_i];
       }
     } else {
       for(i=0;i<count;i++) {
	 prtl = temp4[i];
	 /* Really shouldn't use ad_GRAD_MAX; should use adaxpy, if possible */
	 for(g_i=0;g_i<ad_GRAD_MAX;g_i++)
	   bigtemp[pos++] = prtl*DERIV_grad(ad_in[i])[g_i];
       }
     }
   }

   ierr = MPI_Reduce(bigtemp,bigtemp2,pos,MPI_DOUBLE,MPI_SUM,root,comm);

   if (myid == root) {
     pos = 0;

     for(i=0;i<count;i++)
       for(g_i=0;g_i<ad_GRAD_MAX;g_i++)
	 DERIV_grad(ad_out[i])[g_i] = bigtemp[pos++];

     ierr = MPI_Recv(temp1,count,MPI_DOUBLE,MPI_ANY_SOURCE,1970,comm,&status);

     for(i=0;i<count;i++)
       DERIV_val(ad_out[i]) = temp1[i];
   }
    } else {
      printf("Unknown reduction operator for type DERIV_TYPE.\n");
      return 1;
    }
  }
}      
