#
# Package: BufferedInput
# 
# Modification history:
# 5 Jan 96: Created
#
# Usage summary:
# \begin{itemize}
# \item [[input_initialize]] ([[*FILE]], [[$interesting_regex]], 
#   [[[$buffer_size]]]): Initialize the package to read from FILE (which
#   you must open and close). The [[$buffer_size]] is optional; it
#   determines the largest number of continuation lines (minus one) a
#   line may have and still be correctly processed. The default is 100.
# 
# \item [[input_get]] ([[*line]]): Put next line in [[$line]]. Return
#   undef if no more lines.
# 
# \item [[input_interesting]]: Returns 1 if the last line ``gotten'' was
#   interesting, and 0 otherwise.
# \end{itemize}
# 

use Exporter;

package BufferedInput;

@ISA = qw(Exporter);
@EXPORT = qw(input_initialize input_get input_interesting input_line_number);
@EXPORT_OK = qw();


#    $cont_line_regex = '^     [^\s0]';
    @input_buffer = ();
# Declare variables used in package.
    $input_buffersize = 100;
    $input_current_size = 0;
    $was_interesting = 0;
#    @processed_buffer = ();
    $line_number = 1;
    $returned_line_number = 1;
sub input_initialize {
    (*FILE, $interesting_line, $input_buffersize) = @_;

    if ( ! defined($input_buffersize) ) {
        $input_buffersize = 100;
    }
    
    @input_buffer = ();
    $was_interesting = 0;
    $input_current_size = 0;
    $line_number = 1;
    $returned_line_number = -1;
}
sub fill_buffer {
    $counter = $input_buffersize - $input_current_size;
    $lines_wanted = $counter;
    
    while (($counter > 0) && ($_ = <FILE>))
    {
        $counter--;
        chop;
        push(@input_buffer, $_);
    }
    $input_current_size += $lines_wanted - $counter;
}
sub input_get {
    local(*_) = @_;

    # < Unget code for get line routine > would go here.
    &fill_buffer;

    $_ = shift(@input_buffer);

    # no more lines?
    if ( ! defined ($_) ) {
        return undef;
    }

    # got a line

    $returned_line_number = $line_number;
    $input_current_size--;
    $line_number++;

    # is interesting? ... no
    if ( ! m/$interesting_line/io ) {
        $was_interesting  = 0;
        return 1;
    }

    # must be interesting...

    $was_interesting = 1;
    return 1;
}
## < < Input: Unget > >
sub input_interesting {
    return $was_interesting;
}
sub input_line_number {
    return $returned_line_number;
}
1;
